from pathlib import Path
from typing import Self, TypedDict, Unpack, override

import anyio
import asyncer
from azure.storage.blob.aio import BlobServiceClient, ContainerClient
from azure.storage.blob.aio._blob_client_async import BlobClient

from azaux.storage_resource_manager import StorageResourceManager


class ServiceKwargs(TypedDict, total=False):
    secondary_hostname: str
    max_block_size: int
    max_single_put_size: int
    min_large_block_upload_threshold: int
    use_byte_buffer: bool
    max_page_size: int
    max_single_get_size: int
    max_chunk_get_size: int
    audience: str


class ContainerManager(StorageResourceManager):
    """Class to manage retrieving blob data from a given blob file.

    :param container: The name of the container.
    :param account: The name of the Azure Storage account.
    :param api_key: The API key for the Azure Storage account.
    """

    def __init__(
        self,
        container: str,
        account: str,
        api_key: str,
        **kwargs: Unpack[ServiceKwargs],
    ) -> None:
        self.container = container
        super().__init__(account, api_key)
        self.kwargs = kwargs
        self._service_client_cache: BlobServiceClient | None = None
        self._client_cache: ContainerClient | None = None

    @override
    async def __aenter__(self) -> Self:
        await self._get_service_client().__aenter__()
        return self

    @override
    async def __aexit__(self, *args: object) -> None:
        await self._get_service_client().__aexit__(*args)
        self._service_client_cache = None
        self._client_cache = None

    @override
    def get_endpoint(self) -> str:
        return self._get_endpoint("blob")

    def _get_service_client(self) -> BlobServiceClient:
        if not self._service_client_cache:
            self._service_client_cache = BlobServiceClient(
                self.get_endpoint(), credential=self.api_key, **self.kwargs
            )
        return self._service_client_cache

    def _get_client(self) -> ContainerClient:
        if not self._client_cache:
            service_client = self._get_service_client()
            self._client_cache = service_client.get_container_client(self.container)
        return self._client_cache

    async def download(self, blob_path: Path) -> bytes:
        """Retrieve data from a given blob file within the container.

        :param blob_path: The path of the blob file.
        :param use_chunks: Whether to use chunking to download the blob.
        :return: The content of the blob.
        """
        downloader = await self._get_client().download_blob(
            str(blob_path), encoding=None
        )
        return await downloader.readall()  # NOTE: to decode bytes.decode(encoding)

    async def download_to_file(self, blob_path: Path, filepath: Path) -> None:
        """Download a blob from the container to a file."""
        async with await anyio.open_file(filepath, mode="wb") as f:
            await f.write(await self.download(blob_path))

    async def upload(self, filepath: Path, blob_path: Path | None = None) -> BlobClient:
        """Upload a file to a blob.

        :param filepath: The path to the file to upload.
        :param blob_path: The path in the blob storage to upload to.
        :return: The BlobClient for the uploaded blob.
        """
        async with await anyio.open_file(filepath, mode="rb") as f:
            blob_name = str(blob_path or filepath)
            return await self._get_client().upload_blob(blob_name, f)

    async def sync_blobs(
        self, filepaths: list[Path], blob_paths: list[Path] | None = None
    ) -> list[BlobClient]:
        """Upload files only if they don't already exist in the blob storage.

        :param filepaths: The paths to the files to upload.
        :param blob_paths: The paths in the blob storage to upload to.
        :return: The BlobClients for the uploaded blobs.
        """
        soon_values: list[asyncer.SoonValue[BlobClient]] = []
        async with asyncer.create_task_group() as tg:
            for i, pth in enumerate(filepaths):
                blb = blob_paths[i] if blob_paths else pth
                blob_client = self._get_client().get_blob_client(blb.as_posix())
                if not await blob_client.exists():
                    soon_values.append(tg.soonify(self.upload)(pth, blb))
        return [sv.value for sv in soon_values]

    async def sync_with_folder(
        self, folder: Path, pattern: str = "**/*"
    ) -> list[BlobClient]:
        """Sync a folder with the blob storage container."""
        existing_blobs = {
            blb.name async for blb in self._get_client().list_blobs(folder.name + "/")
        }
        tasks: list[asyncer.SoonValue[BlobClient]] = []
        async with asyncer.create_task_group() as tg:
            for pth in folder.glob(pattern):
                if pth.is_file():
                    blb_path = pth.relative_to(folder.parent)
                    if str(blb_path) not in existing_blobs:
                        tasks.append(tg.soonify(self.upload)(pth, blb_path))
        return [t.value for t in tasks]



from typing import Self, override

from azure.data.tables import TableEntity
from azure.data.tables.aio import TableClient, TableServiceClient

from azaux.storage_resource_manager import StorageResourceManager


class TableManager(StorageResourceManager):
    """Class to manage table entities from a given table from the Table Storage."""

    def __init__(
        self,
        table: str,
        account: str,
        api_key: str,
    ) -> None:
        self.table = table
        super().__init__(account, api_key)
        self._service_client_cache: TableServiceClient | None = None
        self._client_cache: TableClient | None = None

    @override
    async def __aenter__(self) -> Self:
        await self._get_service_client().__aenter__()
        return self

    @override
    async def __aexit__(self, *args: object) -> None:
        await self._get_service_client().__aexit__(*args)
        self._service_client_cache = None
        self._client_cache = None

    def _get_service_client(self) -> TableServiceClient:
        if not self._service_client_cache:
            self._service_client_cache = TableServiceClient(
                self.get_endpoint(), credential=self.get_creds()
            )
        return self._service_client_cache

    def _get_client(self) -> TableClient:
        if not self._client_cache:
            self._client_cache = self._get_service_client().get_table_client(self.table)
        return self._client_cache

    @override
    def get_endpoint(self) -> str:
        return self._get_endpoint("table")

    async def upsert_entity(self, entity_data: dict) -> None:
        """Upload a table entity to the table storage account.

        :param entity_data: The data to be uploaded as a table entity.
        """
        await self._get_client().upsert_entity(entity=entity_data)

    async def retrieve_table_entities(self, query: str) -> list[TableEntity]:
        """Retrieve table entities from the table storage account.

        :param query: The query string to filter the table entities.
        """
        client = self._get_client()
        return [ent async for ent in client.query_entities(query_filter=str(query))]

    async def remove_table_entity(self, entity: TableEntity) -> None:
        """Remove a table entity from the table storage account.

        :param entity: The table entity to be removed.
        """
        await self._get_client().delete_entity(
            partition_key=entity["PartitionKey"], row_key=entity["RowKey"]
        )



import re
from abc import ABC, abstractmethod
from typing import Literal, Self

from azure.core.credentials import AzureNamedKeyCredential


class StorageResourceManager(ABC):
    """Abstract class for managing storage resources in Azure Storage.

    :param account: The name of the Azure Storage account.
    :param api_key: The API key for the Azure Storage account.
    """

    class InvalidConnectionStringError(ValueError):
        """Exception raised when an Azure Storage connection string is invalid."""

        def __init__(self, connection_string: str) -> None:
            """Initialize the InvalidConnectionStringError class."""
            self.connection_string = connection_string
            super().__init__(
                f"Invalid Azure Storage connection string: {connection_string}"
            )

    def __init__(self, account: str, api_key: str) -> None:
        """Initialize the StorageResourceManager class."""
        self.account = account
        self.api_key = api_key

    @abstractmethod
    async def __aenter__(self) -> Self:
        pass

    @abstractmethod
    async def __aexit__(self, *args: object) -> None:
        pass

    def get_creds(self) -> AzureNamedKeyCredential:
        """Return the credentials for the storage resource."""
        return AzureNamedKeyCredential(self.account, self.api_key)

    def _get_endpoint(self, resource_type: Literal["blob", "queue", "table"]) -> str:
        """Return the endpoint URL for the storage resource."""
        return f"https://{self.account}.{resource_type}.core.windows.net"

    @abstractmethod
    def get_endpoint(self) -> str:
        """Return the endpoint URL for the storage resource."""

    @classmethod
    def from_connection_string(cls, connection_string: str) -> Self:
        """Create a new instance from an Azure Storage connection string.

        :param connection_string: The Azure Storage connection string.
        :return: A new instance of the AzureNamedKeyCredential class.
        :raise InvalidConnectionStringError: If the connection string is invalid.
        """
        pattern = r"DefaultEndpointsProtocol=(.*);AccountName=(.*);AccountKey=(.*);"
        match = re.match(pattern, connection_string)
        if not match:
            raise cls.InvalidConnectionStringError(connection_string)
        _, account, key = match.groups()
        return cls(account, key)



from typing import Self, override

import asyncer
from azure.storage.queue.aio import QueueClient, QueueServiceClient

from azaux.storage_resource_manager import StorageResourceManager


class QueueManager(StorageResourceManager):
    """Class to manage sending messages to a given queue from the Queue Storage account.

    :param queue: The name of the queue.
    :param account: The name of the Azure Storage account.
    :param api_key: The API key for the Azure Storage account.
    """

    def __init__(self, queue: str, account: str, api_key: str) -> None:
        self.queue = queue
        super().__init__(account, api_key)
        self._service_client_cache: QueueServiceClient | None = None
        self._client_cache: QueueClient | None = None

    @override
    async def __aenter__(self) -> Self:
        await self._get_service_client().__aenter__()
        return self

    @override
    async def __aexit__(self, *args: object) -> None:
        await self._get_service_client().__aexit__(*args)
        self._service_client_cache = None
        self._client_cache = None

    @override
    def get_endpoint(self) -> str:
        return self._get_endpoint("queue")

    def _get_service_client(self) -> QueueServiceClient:
        if not self._service_client_cache:
            self._service_client_cache = QueueServiceClient(
                self.get_endpoint(), credential=self.api_key
            )
        return self._service_client_cache

    def _get_client(self) -> QueueClient:
        if not self._client_cache:
            self._client_cache = self._get_service_client().get_queue_client(self.queue)
        return self._client_cache

    async def send_messages(self, instance_inputs: list[str]) -> None:
        """Send messages to the queue.

        :param instance_inputs: The list of messages to send to the queue.
        """
        async with asyncer.create_task_group() as tg:
            for input_msg in instance_inputs:
                tg.soonify(self._get_client().send_message)(input_msg)
